# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.


__version__ = "0.7.1"


__all__ = [
    '__version__'
]
