# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.


class ParserError(RuntimeError):
    pass


class ParserWarning(RuntimeWarning):
    pass
